/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Vidur Apparao <vidur@netscape.com> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "nsISupports.idl"
#include "nsIWSDL.idl"

[scriptable, uuid(bb76b72e-d143-4738-804a-53b311466dd9)]
interface nsIWSDLSOAPBinding : nsIWSDLBinding
{
  const unsigned short STYLE_RPC = 1;
  const unsigned short STYLE_DOCUMENT = 2;
};

[scriptable, uuid(0458dac7-65de-11d5-9b42-00104bdf5339)]
interface nsISOAPPortBinding : nsIWSDLSOAPBinding
{
  const unsigned short SOAP_VERSION_1_1 = 0;
  const unsigned short SOAP_VERSION_1_2 = 1;
  const unsigned short SOAP_VERSION_UNKNOWN = 32767;

  readonly attribute AString name;
  readonly attribute AString address;
  readonly attribute unsigned short style;
  readonly attribute AString transport;
  readonly attribute unsigned short soapVersion;
};

[scriptable, uuid(0458dac8-65de-11d5-9b42-00104bdf5339)]
interface nsISOAPOperationBinding : nsIWSDLSOAPBinding
{
  readonly attribute unsigned short style;
  readonly attribute AString soapAction;
};

[scriptable, uuid(c3c4614c-301a-44af-ad70-936f9fd1ba5c)]
interface nsISOAPMessageBinding : nsIWSDLSOAPBinding
{
  readonly attribute AString namespace;
};

[scriptable, uuid(0458dac9-65de-11d5-9b42-00104bdf5339)]
interface nsISOAPPartBinding : nsIWSDLSOAPBinding
{
  const unsigned short   LOCATION_BODY = 1;
  const unsigned short   LOCATION_HEADER = 2;
  const unsigned short   LOCATION_FAULT = 3;

  const unsigned short   USE_LITERAL = 1;
  const unsigned short   USE_ENCODED = 2;

  readonly attribute unsigned short location;
  readonly attribute unsigned short use;
  readonly attribute AString encodingStyle;
  readonly attribute AString namespace;
};
